#import "BuiltInConsole.h"
#import "RInterpreter.h"

@implementation BuiltInConsole


- (IBAction)sourceableChecked:(id)sender {
	switch([sender state]) {
		case NSOnState:
			[comments setEnabled:YES];
		break;
		case NSOffState:
			[comments setEnabled:NO];
		break;
	}
}


- (BOOL)prepareSavePanel:(NSSavePanel*)aSavePanel {
	[aSavePanel setAccessoryView:savePanel];
	return YES;
}

- (NSString *)windowNibName {
    return @"BuiltInConsole";
}

- (void)makeWindowControllers {
	//Load our save panel
	[NSBundle loadNibNamed:@"SaveAuxView" owner:self];
	//Set the window controllers
	ctrl = [[BuiltInConsoleController alloc] initWithWindowNibName:[self windowNibName]];
	[self addWindowController:ctrl];
}

- (NSDictionary *)fileAttributesToWriteToURL:(NSURL *)absoluteURL ofType:(NSString *)typeName forSaveOperation:(NSSaveOperationType)saveOperation originalContentsURL:(NSURL *)absoluteOriginalContentsURL error:(NSError **)outError {
	NSMutableDictionary *d2 = [NSMutableDictionary dictionaryWithDictionary:[super fileAttributesToWriteToURL:absoluteURL ofType:typeName forSaveOperation:saveOperation originalContentsURL:absoluteOriginalContentsURL error:outError]];
	if(nil == typeName) {
		long *lp     = (long*)"crsR";
		[d2 setObject:[NSNumber numberWithUnsignedLong:*lp] forKey:NSFileHFSTypeCode];
		[d2 setObject:[NSNumber numberWithUnsignedLong:*lp] forKey:NSFileHFSCreatorCode];		
	}
	NSLog([d2 description]);
	return d2;
}
- (NSData *)dataOfType:(NSString *)type error:(NSError**)error; {
	
	BOOL source = ([sourceable state] == NSOnState) ? YES : NO;
	BOOL comment= (source && [comments state] == NSOnState) ? YES : NO;
	NSTextStorage *store = [ctrl textStorage];
	NSString *text = [store string];
	
	NSMutableString *script = [[NSMutableString alloc] init];
	
	NSInteger pos   = 0,len = [text length];
	NSLog(type);
	while(pos < len) {
		NSRange range;
		NSString *textType = [store attribute:@"RTextType" atIndex:pos effectiveRange:&range];
		NSLog(@"%@ [%@]",textType,[text substringWithRange:range]);
		if([textType isEqual:@"OutputText"]) {
			//If we don't have a comment we can just append the string. Otherwise we need to insert # as the first character of all lines.
			if(NO == comment) {
				//If this IS sourceable and NOT commented then don't print it (otherwise it won't be sourceable!)
				if(NO == source) [script appendString:[text substringWithRange:range]];
			} else {
				for(NSString *line in [[text substringWithRange:range] componentsSeparatedByString:@"\n"]) {
					[script appendFormat:@"#%@\n",line];
				}
			}
		} else if(NO == source && [textType isEqual:@"PromptText"]) {
			[script appendString:[text substringWithRange:range]];
		} else if([textType isEqual:@"InputText"]) {
			[script appendString:[text substringWithRange:range]];
		}
		pos = range.location + range.length;
	}
	*error = nil;
	return [script dataUsingEncoding:NSUTF8StringEncoding allowLossyConversion:YES];
}

- (void)awakeFromNib {
	
	//[[consoleText textStorage] setDelegate:self];
	
	//Watch console changes so that will know when we're active.
}


@end
