#import <Cocoa/Cocoa.h>
#import "RGraphicsDeviceDelegateProxy.h"

//The RGraphicsDevice is not a general wrapper around
//the R graphics device per se. Rather, it is a wrapper for
//managing a particular device---the CoreGraphics device that
//is used to manage all of the R drawing operations under OS X.
//Basically, if you can get a CGContextRef for it, we can draw to it.

//Leopard-specific routines (particularly for text) are employed.

@interface RGraphicsDevice : NSObject {
	BOOL			   open;
	NSInteger          deviceNumber;
	CGContextRef	   context;
	RGraphicsDeviceDelegateProxy *delegate;
	void* devDesc;
	
	double  scaleFactor,pointSize,textScale;
	
	BOOL    locateSet;
	NSPoint locatePoint;
}

@property(readwrite) double    scaleFactor;
@property(readwrite) double    pointSize;
@property(readwrite) double    textScale;
@property(readonly)  NSInteger deviceNumber;


+ (RGraphicsDevice*)deviceWithContext:(CGContextRef)aContext deviceDesc:(void*)dd delegate:(id)aDelegate;
- (id)initWithContext:(CGContextRef)aContext deviceDesc:(void*)dd delegate:(id)aDelegate;

- (CGContextRef)context;
- (void)setContext:(CGContextRef)aContext;

- (void)setSize:(NSSize)size;

- (void)beginUpdate;
- (void)endUpdate;

- (NSSize)deviceSize;
- (void)newPage;
- (void)closeDevice;

- (void)redraw;


- (BOOL)isLocating;
- (void)setLocatePoint:(NSPoint)p;
- (void)stopLocating;


@end
