#import "RInterpreter.h"
#import "RValue.h"

#define R_INTERFACE_PTRS 1
#define CSTACK_DEFNS     1

#include <Rinternals.h>
#include <Rinterface.h>
#include <R_ext/Utils.h>
#include <Rgraphics.h>
#include <R_ext/GraphicsDevice.h>
#include <R_ext/eventloop.h>

#undef Boolean
//This number seems to work well. Vast quantities of output are still
//much slower than via Terminal.app however (<1sec vs ~4.3sec). I wonder if there is some way of causing layout to be less frequent when printing output.
#define INTERP_BUFFER 2048 

NSString *RInterpreterWillWriteToConsole = @"RInterpreterWillWriteToConsole";
NSString *RInterpreterDidWriteToConsole  = @"RInterpreterDidWriteToConsole";

NSString *RInterpreterWillWritePrompt = @"RInterpreterWillWritePrompt";
NSString *RInterpreterDidWritePrompt  = @"RInterpreterDidWritePrompt";

NSString *RInterpreterWillReadFromConsole = @"RInterpreterWillReadFromConsole";
NSString *RInterpreterDidReadFromConsole = @"RInterpreterDidReadFromConsole";

NSString *RInterpreterDidFinishStartingInterpreter = @"RInterpreterDidFinishStartingInterpreter";
NSString *RInterpreterWillStartEvaluating = @"RInterpreterWillStartEvaluating";
NSString *RInterpreterDidStartEvaluating = @"RInterpreterDidStartEvaluating";
NSString *RInterpreterDidFinishEvaluating = @"RInterpreterDidFinishEvaluating";

NSString *RInterpreterWillChangeConsole = @"RInterpreterWillChangeConsole";
NSString *RInterpreterDidChangeConsole  = @"RInterpreterDidChangeConsole";

NSString *RInterpreterShowHelpForFunction = @"RInterpreterShowHelpForFunction";

NSString *RInterpreterOpenDevice = @"RInterpreterOpenDevice";
NSString *RInterpreterShowHelpFiles = @"RInterpreterShowHelpFiles";


#pragma mark Global Variables
#pragma mark -
RInterpreter *RInterp = nil;
int  (*ptr_R_OldReadConsole)(char*,unsigned char*,int,int) = NULL;
BOOL suppressOutput   = NO;
BOOL currentHistory   = NO;

//For ESS support
NSLock   *thread_lock = nil;
char *thread_aPrompt;
unsigned char *thread_aBuffer;
NSInteger thread_aLength;
NSInteger thread_addToHistory;
NSInteger thread_result;


#pragma mark Console Interface
CFMutableAttributedStringRef consoleOutput = NULL;
CFMutableStringRef           consoleBuffer = NULL;
CFMutableStringRef           consoleInput  = NULL;

#pragma mark Console Input/Output Attributes
BOOL                         updateAttrs   = YES;
CFMutableDictionaryRef       promptAttrs   = NULL;
CFMutableDictionaryRef       outputAttrs   = NULL;
CFMutableDictionaryRef       inputAttrs    = NULL;

#pragma mark State Flags
Boolean firstPrompt = true;
Boolean firstConsole= true;
char   *currentPrompt;

#pragma mark Helper Functions 
#pragma mark -
//A helper function that doesn't exist by default.
void CFAttributedStringAppendAttributedString(CFMutableAttributedStringRef aStr,CFAttributedStringRef toAppend) {
	CFAttributedStringReplaceAttributedString(aStr,CFRangeMake(CFAttributedStringGetLength(aStr), 0),toAppend);
}
void CFAttributedStringAppendString(CFMutableAttributedStringRef aStr,CFStringRef toAppend,CFDictionaryRef attrs) {
	CFAttributedStringRef attr = CFAttributedStringCreate(kCFAllocatorDefault,toAppend,attrs);
	CFAttributedStringAppendAttributedString(aStr,attr);
	CFRelease(attr);
}

void RInterpConsoleAppendString(CFStringRef aStr) {
	CFStringAppend(consoleInput,aStr);
}
void RInterpConsoleAppendCString(char *aStr) {
	CFStringAppendCString(consoleInput,aStr,kCFStringEncodingUTF8);
}

void CopyConsoleWithNewAttributes(CFMutableAttributedStringRef aStr,CFAttributedStringRef toAppend) {
	CFStringRef justStr = CFAttributedStringGetString(toAppend);
	CFRange work = CFRangeMake(0,CFStringGetLength(justStr));
	while(work.length > 0) {
		CFRange     effRange;
		CFStringRef chunkType = CFAttributedStringGetAttributeAndLongestEffectiveRange(
			toAppend,work.location,CFSTR("RTextType"),work,&effRange);
		CFDictionaryRef attrs;
		if(CFStringCompare(chunkType,CFSTR("PromptText"),0)) {
			attrs = promptAttrs;
		} else if(CFStringCompare(chunkType,CFSTR("InputText"),0)) {
			attrs = inputAttrs;
		} else if(CFStringCompare(chunkType,CFSTR("OutputText"),0)) {
			attrs = outputAttrs;
		}
		//Adjust possible range behind me.
		if(effRange.location < work.location)
			effRange.length -= work.location-effRange.location;
		effRange.location = work.location;

		CFStringRef tmp = CFStringCreateWithSubstring(kCFAllocatorDefault,justStr,effRange);
		CFAttributedStringAppendString(aStr,tmp,attrs);
		CFRelease(tmp);
		work.location += effRange.length;
		work.length   -= effRange.length;
	}
	CFRelease(justStr);
}



#pragma mark Function Prototypes
void RInterp_Suicide(char*);
void RInterp_ShowMessage(char*);
void RInterp_FlushConsole();
void RInterp_WritePrompt(char*);
int  RInterp_ReadConsole(char*,unsigned char*,int,int);
int  RInterp_ThreadedReadConsole(char*,unsigned char*,int,int);
void RInterp_ResetConsole();
void RInterp_WriteConsole(char*,int);
void RInterp_ClearerrConsole();
void RInterp_Busy();
void RInterp_CleanUp(SA_TYPE,int,int);
int  RInterp_ShowFiles(int,char**,char**,char*,Rboolean,char*);
int  RInterp_ChooseFile(int,char*,int);
int  RInterp_EditFile(char*);
void RInterp_System(char*);
void RInterp_ProcessEvents();
Rboolean RInterp_Device(NewDevDesc*,char*,double,double,double,char*,Rboolean,Rboolean,int,int);
void RInterp_DeviceParams(double*,double*,double*,char*,Rboolean*,Rboolean*,int*);
int  RInterp_CustomPrint(char *,SEXP);



//Who would ever want access to R_ProcessEvents? Oh, right. Anyone writing a front end.
extern void     (*ptr_R_ProcessEvents)(void);
extern void     (*ptr_CocoaSystem)(char*);
extern Rboolean (*ptr_CocoaInnerQuartzDevice)(NewDevDesc*,char*,double,double,double,char*,Rboolean,Rboolean,int,int);
extern void     (*ptr_CocoaGetQuartzParameters)(double*,double*,double*,char*,Rboolean*,Rboolean*,int*);
extern int      (*ptr_Raqua_CustomPrint)(char *, SEXP);


@interface RInterpreter (Private)
- (void)processEvents;

- (void)willWriteToConsole;
- (void)didWriteToConsole;

- (void)evaluateInput;

- (void)_threadedReader:(id)sender;

@end


@implementation RInterpreter
+ (RInterpreter*)sharedInterpreter {
	if(nil == RInterp) RInterp = [[self alloc] init];
	return RInterp;
}

- (id)init {
	if(nil == [super init]) return nil;

	//By default we are in R.app mode.
	standardReader = NO;
	
	//Create (temporary) console outpt and console input locations
	consoleOutput = CFAttributedStringCreateMutable(kCFAllocatorDefault,0);
	consoleInput  = CFStringCreateMutable(kCFAllocatorDefault,0);
	consoleBuffer = CFStringCreateMutable(kCFAllocatorDefault,INTERP_BUFFER);

	//Create dictionaries for the input and output attributes
	promptAttrs = CFDictionaryCreateMutable(kCFAllocatorDefault,0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
	CFDictionaryAddValue(promptAttrs,CFSTR("RTextType"),CFSTR("PromptText"));
	outputAttrs = CFDictionaryCreateMutable(kCFAllocatorDefault,0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
	CFDictionaryAddValue(outputAttrs,CFSTR("RTextType"),CFSTR("OutputText"));
	inputAttrs = CFDictionaryCreateMutable(kCFAllocatorDefault,0, &kCFTypeDictionaryKeyCallBacks, &kCFTypeDictionaryValueCallBacks);
	CFDictionaryAddValue(inputAttrs,CFSTR("RTextType"),CFSTR("InputText"));
	
	//Library paths
	libraryPaths  = [[NSMutableArray alloc] init];
	
	
	//Locale information. For some reason, my machine thinks that it should be of a Japanese locale (LANG is set...). Weird, eh?
	//Anyway, override this with whatever language Cocoa applications think they should be since this is what the NIBs will be 
	//using for their locale information. On the bright side the Japanese text rendered perfectly! 
	if(nil != [[NSUserDefaults standardUserDefaults] stringForKey:@"DefaultLocaleString"])
		setenv("LANG",[[[NSUserDefaults standardUserDefaults] stringForKey:@"DefaultLocaleString"] UTF8String],1);
	else {
		[NSString stringWithFormat:@"%@.UTF-8",[[NSLocale currentLocale] localeIdentifier]];
		setenv("LANG",[[[NSLocale currentLocale] localeIdentifier] UTF8String],1);
	}
	
	periodic = nil;
	
	return self;
}

- (void)setPromptAttributes:(NSDictionary*)attrs {
	NSMutableDictionary *dict = (NSMutableDictionary*)promptAttrs;
	[dict removeAllObjects];
	if(nil != attrs)
		[dict addEntriesFromDictionary:attrs];
	CFDictionaryAddValue(promptAttrs,CFSTR("RTextType"),CFSTR("PromptText"));
}
- (void)setOutputAttributes:(NSDictionary*)attrs {
	NSMutableDictionary *dict = (NSMutableDictionary*)outputAttrs;
	[dict removeAllObjects];
	if(nil != attrs)
		[dict addEntriesFromDictionary:attrs];
	CFDictionaryAddValue(outputAttrs,CFSTR("RTextType"),CFSTR("OutputText"));
}
- (void)setInputAttributes:(NSDictionary*)attrs {
	NSMutableDictionary *dict = (NSMutableDictionary*)inputAttrs;
	[dict removeAllObjects];
	if(nil != attrs)
		[dict addEntriesFromDictionary:attrs];
	CFDictionaryAddValue(inputAttrs,CFSTR("RTextType"),CFSTR("InputText"));
}
- (NSDictionary*)inputAttributes {
	return (NSDictionary*)inputAttrs;
}


- (void)setUsesStandardReader:(BOOL)aBool {
	standardReader = aBool;
}
- (BOOL)usesStandardReader {
	return standardReader;
}

- (void)setSuppressOutput:(BOOL)aBool {
	suppressOutput = aBool;
}
- (BOOL)suppressOutput { return suppressOutput; }


extern  void Rstd_WriteConsole(char *buf, int len);

- (void)run {
	char *argv[] = {"R","--gui=cocoa","--no-save","--no-restore-data",0};
	
	
	[[RValue gil] lock];
	Rf_initialize_R(4,argv);
	R_Interactive     = 1;
	//If we aren't using the standard reader, take over everything.
	//Otherwise, just swizzle the reader for multithreading. 
	if(NO == standardReader) {
		ptr_R_Suicide     = RInterp_Suicide;
		ptr_R_ShowMessage = RInterp_ShowMessage;
		ptr_R_ReadConsole = RInterp_ReadConsole;
		ptr_R_WriteConsole= RInterp_WriteConsole;
		ptr_R_ShowFiles   = RInterp_ShowFiles;
		ptr_R_EditFile    = RInterp_EditFile;
		ptr_R_ChooseFile  = RInterp_ChooseFile;
		ptr_Raqua_CustomPrint = RInterp_CustomPrint;

		R_Outputfile = NULL;
		R_Consolefile = NULL;
	} else {
		ptr_R_WriteConsole   = Rstd_WriteConsole;
		ptr_R_OldReadConsole = ptr_R_ReadConsole;
		ptr_R_ReadConsole    = RInterp_ThreadedReadConsole;
		//Restore these.
		R_Outputfile = stdout;
		R_Consolefile = stderr;
	}
	ptr_R_ProcessEvents          = RInterp_ProcessEvents;
	ptr_CocoaInnerQuartzDevice   = RInterp_Device;
	ptr_CocoaGetQuartzParameters = RInterp_DeviceParams;
	ptr_CocoaSystem              = RInterp_System;
	
	setup_Rmainloop();
	[RInterp didFinishStartingInterpreter];
	run_Rmainloop();
}

- (void)stop {
	[self evaluateInput];
}

- (void)setHomeDirectory:(NSString*)homeDir {
	setenv("R_HOME",[homeDir UTF8String],1);
}

- (BOOL)shouldAddToHistory { return currentHistory; }

- (void)appendInput:(NSString*)aStr evaluate:(BOOL)shouldEvaluate {
	CFStringAppend(consoleInput,(CFStringRef)aStr);
	if(YES == shouldEvaluate)
		[self evaluateInput];
}

- (BOOL)hasActiveConsole { return true == firstConsole ? NO : YES; }

- (void)setConsoleDocument:(id)aConsole buffer:(CFMutableAttributedStringRef)aBuffer copy:(BOOL)shouldCopy {
	if(aBuffer == consoleOutput) return;
	if(RInterp_FlushConsole_ret() > 0)
		[self didWriteToConsole];
	
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterWillChangeConsole object:aConsole]];		
	if(NULL != consoleOutput && YES == shouldCopy) {
		CFAttributedStringBeginEditing(aBuffer);
		CopyConsoleWithNewAttributes(aBuffer,consoleOutput);
		CFAttributedStringEndEditing(aBuffer);
	} 
	CFMutableAttributedStringRef oldConsole = consoleOutput;
	consoleOutput = aBuffer;
	CFRetain(aBuffer);
	CFRelease(oldConsole);
	
	//This is no longer the first console
	firstConsole = false;
	//If we didn't copy the console text and we are at a prompt,
	//we should send the prompt to the console so it knows where input should start.
	if(NULL != currentPrompt && NO == shouldCopy)
		RInterp_WritePrompt(currentPrompt);
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterDidChangeConsole object:aConsole]];		
}

@end


@implementation RInterpreter (Private)
- (void)processEvents {
	//TODO: Implement R_ProcessEvents coalescing since it can be called QUITE often. 
	//We can sacrifice some interactivity for faster R processing.
	if(NULL != R_PolledEvents) {
		R_PolledEvents();	
	}
	[self stop]; //Mark the event queue
	[NSApp run]; //Process events in the queue
}

- (void)periodicEvents:(id)userInfo {
	NSLog(@"periodic event");
	if(NULL != R_PolledEvents) R_PolledEvents();
}

- (void)suicide:(NSString*)aMsg {
	NSLog(@"Suicide: %@",aMsg);
}
- (void)showMessage:(NSString*)aMsg {
	NSLog(@"ShowMessage: %@",aMsg);
}
- (void)waitForInput {
	[[RValue gil] unlock];
	[NSApp run];
	[[RValue gil] lock];
}

- (void)evaluateInput {
	[NSApp postEvent:[NSEvent otherEventWithType:NSApplicationDefined 
									 location:NSMakePoint(0,0)
								modifierFlags:0 
									timestamp:0 
								 windowNumber:0 
									  context:nil 
									  subtype:1337
										data1:0 
										data2:0] atStart:NO];
}

- (void)_threadedReader:(id)sender {
	[thread_lock lock];
	thread_result = ptr_R_OldReadConsole("",thread_aBuffer,thread_aLength,thread_addToHistory);
	[thread_lock unlock];		
	[RInterp performSelectorOnMainThread:@selector(evaluateInput) withObject:nil waitUntilDone:NO];
}

- (void)didWriteToConsole {
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterDidWriteToConsole object:self]];
}
- (void)willWriteToConsole {
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterWillWriteToConsole object:self]];
}
- (void)willWritePrompt {
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterWillWritePrompt object:self]];
}
- (void)didWritePrompt {
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterDidWritePrompt object:self]];
}
- (void)didFinishStartingInterpreter {
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterDidFinishStartingInterpreter object:self]];
}
- (void)willStartEvaluating {
	//Stop sending tcl/tk periodic events (we track them in ProcessEvents)
	if(nil != periodic) {
		[periodic invalidate];
		periodic = nil;
	}
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterWillStartEvaluating object:self]];
}
- (void)didStartEvaluating {
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterDidStartEvaluating object:self]];
}
- (void)didFinishEvaluating {
	//Start sending tcl/tk periodic events 
	if(NULL != R_PolledEvents) {
		periodic = [NSTimer timerWithTimeInterval:0.5
									 target:self
								  selector:@selector(periodicEvents:)
								  userInfo:nil
								   repeats:YES];	
	}
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterDidFinishEvaluating object:self]];
}

- (void)willWaitForInput {
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterWillReadFromConsole object:self]];	
}

- (void)didWaitForInput {
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterDidReadFromConsole object:self]];	
}

@end

#pragma mark C Pointer Routines
#pragma mark -
void RInterp_ProcessEvents() {
	[RInterp processEvents];
}

void RInterp_Suicide(char *aMsg) {
	[RInterp suicide:[NSString stringWithUTF8String:aMsg]];
}

void RInterp_ShowMessage(char *aMsg) {
	[RInterp showMessage:[NSString stringWithUTF8String:aMsg]];
}

NSInteger RInterp_FlushConsole_ret() {
	if(NULL == consoleOutput) return 0;
	NSInteger len = CFStringGetLength(consoleBuffer);
	if(len > 0) {
		CFAttributedStringAppendString(consoleOutput,consoleBuffer,outputAttrs);
		CFStringDelete(consoleBuffer,(CFRange){0,len});
	}
	return len;
}

void RInterp_FlushConsole() {
	RInterp_FlushConsole_ret();
}
void RInterp_WritePrompt(char *aPrompt) {
	if(NULL == consoleOutput || YES == suppressOutput) return;
	[RInterp willWritePrompt];
	CFStringRef prompt = CFStringCreateWithCStringNoCopy(kCFAllocatorDefault,aPrompt,kCFStringEncodingUTF8,kCFAllocatorNull);
	CFAttributedStringAppendString(consoleOutput,prompt,promptAttrs);
	CFRelease(prompt);
	[RInterp didWritePrompt];
}

NSInteger RInterp_ReadConsole(char *aPrompt,unsigned char *aBuffer,NSInteger aLength,NSInteger addToHistory) {
	NSInteger myLength;
	if((myLength = CFStringGetLength(consoleInput)) > 0) {
sendBuffer:
		if(myLength > aLength)
			myLength = aLength;
		CFStringGetCString(consoleInput,(char*)aBuffer,aLength,kCFStringEncodingUTF8);
		aBuffer[myLength] = '\0';
		CFStringDelete(consoleInput,(CFRange){0,myLength});
		return myLength;
	}
	
	//Don't want to finish evaluating on the first pass.
	if(false == firstPrompt) {
		[RInterp didFinishEvaluating];
	} else firstPrompt = false;
	
	currentPrompt = aPrompt;
	RInterp_FlushConsole();
	[RInterp didWriteToConsole];
	currentHistory = addToHistory ? YES : NO;
	RInterp_WritePrompt(aPrompt);
	//We need to set the attributes properly.
	CFRange pos  = {0,0};
	pos.location = CFAttributedStringGetLength(consoleOutput);
	CFAttributedStringSetAttributes(consoleOutput,pos,inputAttrs,true);
	[RInterp willWaitForInput];
	while(0==CFStringGetLength(consoleInput)) [RInterp waitForInput];
	[RInterp didWaitForInput];
	[RInterp willStartEvaluating];
	[RInterp willWriteToConsole];
	myLength = CFStringGetLength(consoleInput);
	goto sendBuffer;
	//We shouldn't ever get here, should we?
	return 0;
}


int RInterp_ThreadedReadConsole(char *aPrompt,unsigned char *aBuffer,int aLength,int addToHistory) {
	//If this is the first time through, init the lock. Make sure we have the lock so the reader
	//doesn't get ahead of itself.
	if(nil == thread_lock) {
		thread_lock = [[NSLock alloc] init];
		[thread_lock lock];
	}
	[NSThread detachNewThreadSelector:@selector(_threadedReader:) toTarget:RInterp withObject:nil];
	//Okay. Stuff to pass to the reader thread.
	thread_aPrompt      = aPrompt;
	thread_aBuffer      = aBuffer;
	thread_aLength      = aLength;
	thread_addToHistory = addToHistory;
	thread_result       = 0;
	//And here we go. 
	if(false == firstPrompt)
		[RInterp didFinishEvaluating];
	else
		firstPrompt = false;
	
	[RInterp didWriteToConsole];
	[RInterp willWritePrompt];
	ptr_R_WriteConsole(aPrompt,strlen(aPrompt));
	[RInterp didWritePrompt];
	[RInterp willWaitForInput];
	[thread_lock unlock];
	[RInterp waitForInput];
	[thread_lock lock];
	[RInterp didWaitForInput];
	[RInterp willStartEvaluating];
	[RInterp willWriteToConsole];	
	return thread_result;
}

void RInterp_ResetConsole() {
}

void RInterp_WriteConsole(char *aBuffer,NSInteger aLength) {
//	printf(aBuffer);
	if(NULL == consoleOutput || YES == suppressOutput) return;
	NSInteger len = CFStringGetLength(consoleBuffer);
	if(aLength > INTERP_BUFFER) {
		RInterp_FlushConsole();
		NSInteger remain = aLength;
		char  *pos = aBuffer;
		while(remain > INTERP_BUFFER) {
			char tmp = pos[INTERP_BUFFER];
			pos[INTERP_BUFFER] = '\0';
			CFStringAppendCString(consoleBuffer,aBuffer,kCFStringEncodingUTF8);
			RInterp_FlushConsole();
			pos[INTERP_BUFFER] = tmp;
			pos += INTERP_BUFFER;
			remain -= INTERP_BUFFER;
		}
		CFStringAppendCString(consoleBuffer,pos,kCFStringEncodingUTF8);
	} else {
		if(len + aLength > INTERP_BUFFER) RInterp_FlushConsole();
		CFStringAppendCString(consoleBuffer,aBuffer,kCFStringEncodingUTF8);
	}
//	CFStringRef buffer = CFStringCreateWithCStringNoCopy(kCFAllocatorDefault,aBuffer,kCFStringEncodingUTF8,kCFAllocatorNull);
//	CFRelease(buffer);
}

void RInterp_ClearerrConsole() {
}

void RInterp_Busy(NSInteger which) {
}

void RInterp_CleanUp(SA_TYPE saveact,NSInteger status,NSInteger runLast) {
}
int RInterp_ShowFiles(NSInteger nfile,char **files,char **headers,char *wtitle,Rboolean del,char *pager) {
	NSLog(@"Should show files");
	return 0;
}
int RInterp_ChooseFile(NSInteger _new,char *buf,NSInteger len) {
	NSLog(@"RInterp_ChooseFile");
	return 0;
}
int RInterp_EditFile(char *filename) {
	NSLog(@"RInterp_EditFile: %s",filename);
	return 0;
}

void RInterp_System(char *command) {
	NSLog(@"system(\"%s\")",command);
}

Rboolean RInterp_Device(NewDevDesc *dd,char *display,
					double width,double height,double pointsize,
					char *family,Rboolean antialias,Rboolean autorefresh,
					int quartzpos,int bg) {
	NSDictionary *dict = [NSDictionary dictionaryWithObjectsAndKeys:
						  [NSValue valueWithPointer:dd],@"NewDevDesc",
						  [NSString stringWithCString:display],@"Display",
						  [NSNumber numberWithDouble:width],@"Width",
						  [NSNumber numberWithDouble:height],@"Height",
						  [NSNumber numberWithDouble:pointsize],@"Pointsize",
						  [NSString stringWithCString:family],@"FontFamily",
						  [NSColor colorWithDeviceRed:R_RED(bg)/255.0 green:R_GREEN(bg)/255.0 blue:R_BLUE(bg)/255.0 alpha:R_ALPHA(bg)/255.0],@"BackgroundColor",
						  nil];
	[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterOpenDevice 
																					  object:dict]];
	NSLog(@"Done trying to create device");
	return (dd->deviceSpecific == NULL) ? FALSE : TRUE;
}
void RInterp_DeviceParams(double *width,double *height,double *ps,char *family,Rboolean *aa,Rboolean *ar,int *pos) {
	//We don't do a thing dammit.
}

int RInterp_CustomPrint(char *type,SEXP args) {
	NSLog(@"CustomPrint: %s",type);
	//We do a bit of unpacking here to help things along
	if(strncmp(type,"help-files",10)==0 && TYPEOF(args) == STRSXP) {
		int i;
		NSMutableArray *all = [[NSMutableArray alloc] init];
		for(i=0;i<length(args);i++)
			[all addObject:[NSString stringWithFormat:@"%s",CHAR(STRING_ELT(args,i))]];
		[[NSNotificationCenter defaultCenter] postNotification:[NSNotification notificationWithName:RInterpreterShowHelpFiles object:all]];
	} else {
		//What do we do for types we don't know? I haven't decided yet.
	}
	return 0;
}


